@@ -10,7 +10,8 @@ import json
 import logging
 import os
 import warnings
-from typing import Any, Dict
+from concurrent.futures import Future, ThreadPoolExecutor
+from typing import Any, Dict, Tuple
 
 import cohere
 import jsonschema
@@ -21,6 +22,7 @@ from conversant.prompts.chat_prompt import ChatPrompt
 from conversant.prompts.prompt import Prompt
 
 MAX_GENERATE_TOKENS = 2048
+TOKENS_PER_REQUEST = 10
 PERSONA_MODEL_DIRECTORY = f"{os.path.dirname(conversant.__file__)}/personas"
 PERSONA_JSON_SCHEMA = {
     "type": "object",
@@ -186,20 +188,54 @@ class PromptChatbot(Chatbot):
             "tokens"
         )
 
-    def reply(self, query: str) -> Interaction:
-        """Replies to a query given a chat history.
+    def _dispatch_concurrent_generate_call(self, **kwargs) -> Future:
+        """Dispatches a concurrent call to co.generate.
 
-        The reply is then generated directly from a call to a LLM.
+        This allows a network bound co.generate call to proceed while also
+        yielding the current response in a partial reply generator.
+
+        Args:
+            kwargs: Keyword arguments for the call to co.generate.
+
+        Returns:
+            Future: A future object that will be called to retrieve the result of
+                co.generate.
+        """
+        with ThreadPoolExecutor(max_workers=1) as exe:
+            future = exe.submit(self.co.generate, **kwargs)
+        return future
+
+    def get_stop_seq(self, response: str) -> str:
+        """Given a response, returns the stop sequence it has if any.
+
+        Args:
+            response (str): Response coming from prompt chatbot.
+
+        Returns:
+            str: The stop sequence in the response. If no stop sequence is found, then
+                an empty string is returned.
+
+        """
+        for stop_seq in self.client_config["stop_sequences"]:
+            if stop_seq in response:
+                return stop_seq
+        return ""
+
+    def generate_prompt_update_examples(self, query: str) -> str:
+        """Generate prompt from query and update max context examples if necessary
 
         Args:
             query (str): A query passed to the prompt chatbot.
 
         Returns:
-            Interaction: Dictionary of query and generated LLM response
+            current_prompt (str): Returns the current prompt using
+            query and chat history
+
         """
         # The current prompt is assembled from the initial prompt,
         # from the chat history with a maximum of max_context_examples,
         # and from the current query
+
         current_prompt = self.get_current_prompt(query)
 
         current_prompt_size = self.co.tokenize(current_prompt).length
@@ -219,6 +255,143 @@ class PromptChatbot(Chatbot):
                 f" to {self.chatbot_config['max_context_examples']} - "
                 f"value set in the original config"
             )
+        return current_prompt
+
+    def partial_reply(self, query: str) -> Tuple[str, str]:
+        """Generates (partial) reply to a query given a chat history.
+
+        Args:
+            query (str): A query passed to the prompt chatbot.
+
+        Yields:
+
+            Tuple[str, str]: A tuple of the response before the co.generate call,
+                and the response after.
+        """
+        current_prompt = self.generate_prompt_update_examples(query)
+        self.prompt_history.append(current_prompt)
+
+        response_before_current = ""
+        response_so_far = ""
+        num_requests_made = 0
+        max_requests = int(self.client_config["max_tokens"] / TOKENS_PER_REQUEST)
+        reply_complete = False
+
+        # As soon as the function is called (and the generator is created), dispatch
+        # a concurrent call to co.generate
+        future = self._dispatch_concurrent_generate_call(
+            model=self.client_config["model"],
+            prompt=current_prompt,
+            max_tokens=TOKENS_PER_REQUEST,
+            temperature=self.client_config["temperature"],
+            frequency_penalty=self.client_config["frequency_penalty"],
+            presence_penalty=self.client_config["presence_penalty"],
+            stop_sequences=self.client_config["stop_sequences"],
+        )
+
+        while num_requests_made < max_requests and not reply_complete:
+            generated_object = future.result()
+            partial_response = generated_object.generations[0].text
+
+            # If the partial response is an empty string, then this iteration is a no-op
+            # (we indicate that the reply is completely generated).
+            if not partial_response:
+                reply_complete = True
+
+            else:
+
+                # Concatenate the candidate response, then fetches the stop sequence if
+                # it exists in the candidate response
+                candidate_response = response_so_far + partial_response
+                stop_seq = self.get_stop_seq(response_so_far + partial_response)
+
+                # Truncate the candidate response if a stop sequence was found
+                if stop_seq:
+                    candidate_response = candidate_response[
+                        : candidate_response.index(stop_seq)
+                    ]
+
+                    # If the stop sequence is found across two partial replies,
+                    # then the response_so_far has to be truncated. Example:
+                    #
+                    #   stop_seq: "\nUser"
+                    #   response_so_far: "Thank you!\n"
+                    #   partial_response: "User: You are welcome"
+                    #
+                    # Then the candidate response is:
+                    #
+                    #   candidate_response: "Thank you!"
+                    #
+                    # In this case, what is yielded at the end of the loop needs to be:
+                    #
+                    #   response_before_current: "Thank you!"
+                    #   response_so_far: "Thank you!"
+                    #
+                    # So we'll truncate the response_so_far to be candidate_response
+                    if len(candidate_response) < len(response_so_far):
+                        response_so_far = candidate_response
+
+                    reply_complete = True
+
+                # Save candidate response
+                current_prompt += partial_response
+                response_before_current = response_so_far
+                response_so_far = candidate_response
+
+                # If this is the first partial_reply, append a new element to
+                # chat history after removing the leading whitespace
+                if num_requests_made == 0:
+                    response_so_far = response_so_far.lstrip()
+                    self.chat_history.append(
+                        self.prompt.create_interaction(query, response_so_far)
+                    )
+                    self.prompt_size_history.append(
+                        self.co.tokenize(
+                            self.prompt.create_interaction_string(
+                                query, response_so_far
+                            )
+                        ).length
+                    )
+                # Otherwise, overwrite the current chat history with the current
+                # response so far
+                else:
+                    self.chat_history[-1] = self.prompt.create_interaction(
+                        query, response_so_far
+                    )
+                    self.prompt_size_history[-1] = self.co.tokenize(
+                        self.prompt.create_interaction_string(query, response_so_far)
+                    ).length
+
+                num_requests_made += 1
+
+                # This dispatches a concurrent call to co.generate, which can be
+                # later accessed on the next iteration of the generator.
+                if num_requests_made < max_requests and not reply_complete:
+                    future = self._dispatch_concurrent_generate_call(
+                        model=self.client_config["model"],
+                        prompt=current_prompt,
+                        max_tokens=TOKENS_PER_REQUEST,
+                        temperature=self.client_config["temperature"],
+                        frequency_penalty=self.client_config["frequency_penalty"],
+                        presence_penalty=self.client_config["presence_penalty"],
+                        stop_sequences=self.client_config["stop_sequences"],
+                    )
+
+                yield response_before_current, response_so_far
+
+    def reply(self, query: str) -> Interaction:
+        """Replies to a query given a chat history.
+
+        The reply is then generated directly from a call to a LLM.
+
+        Args:
+            query (str): A query passed to the prompt chatbot.
+
+        Returns:
+            Interaction: Dictionary of query and generated LLM response
+        """
+
+        current_prompt = self.generate_prompt_update_examples(query)
 
         # Make a call to Cohere's co.generate API
         generated_object = self.co.generate(
@@ -280,10 +453,12 @@ class PromptChatbot(Chatbot):
             context_prompt_lines.append(self.prompt.create_interaction_string(**turn))
         context_prompt = self.prompt.example_separator + "".join(context_prompt_lines)
 
-        # get query prompt
-        query_prompt = self.prompt.create_interaction_string(query)
+        current_prompt = base_prompt + context_prompt
 
-        current_prompt = base_prompt + context_prompt + query_prompt
+        # get query prompt
+        if query != "":
+            query_prompt = self.prompt.create_interaction_string(query)
+            current_prompt += query_prompt
         return current_prompt.strip()
 
     def configure_chatbot(self, chatbot_config: Dict = {}) -> None:
@@ -315,11 +490,11 @@ class PromptChatbot(Chatbot):
         if not hasattr(self, "client_config"):
             self.client_config = {
                 "model": "xlarge",
-                "max_tokens": 100,
+                "max_tokens": 200,
                 "temperature": 0.75,
                 "frequency_penalty": 0.0,
                 "presence_penalty": 0.0,
-                "stop_sequences": ["\n"],
+                "stop_sequences": ["\\n", "\n"],
             }
         # Override default config values with the config passed in
         if isinstance(client_config, Dict):
@@ -368,7 +543,6 @@ class PromptChatbot(Chatbot):
 
         # Validate that the persona follows our predefined schema
         cls._validate_persona_dict(persona, persona_path)
-
         return cls(
             client=client,
             prompt=ChatPrompt.from_dict(persona["chat_prompt_config"]),
